import subprocess
import sys
from pathlib import Path

import click

from sphinx_ape._utils import get_package_name
from sphinx_ape.build import BuildMode, DocumentationBuilder
from sphinx_ape.exceptions import ApeDocsBuildError, ApeDocsPublishError, ApeDocsTestError
from sphinx_ape.testing import DocumentationTester


@click.group()
def cli():
    """
    Documentation CLI
    """


def build_mode_option():
    return click.option(
        "--mode",
        callback=lambda c, p, v: BuildMode.init(v),
        default=BuildMode.LATEST,
    )


def package_name_option():
    return click.option(
        "--name",
        "package_name",
        callback=lambda c, p, v: v or get_package_name(),
    )


@cli.command()
@click.argument("base_path", type=Path)
def init(base_path):
    """
    Initialize documentation structure
    """
    # For this command, force the user to be in that dir.
    builder = _create_builder(base_path=base_path)
    builder.init()


@cli.command()
@click.argument("base_path", type=Path)
@build_mode_option()
@package_name_option()
def build(base_path, mode, package_name):
    """
    Build an Ape python package's documentation
    """
    builder = _create_builder(mode=mode, base_path=base_path, name=package_name)

    # Ensure all the files that need to exist do exist.
    builder.init()

    if not builder.docs_path.is_dir():
        click.echo("docs/ folder missing. Try running:\n\tsphinx-ape init", err=True)
        sys.exit(1)

    click.echo(f"Building '{package_name}' '{mode.name}'.")
    try:
        builder.build()
    except ApeDocsBuildError as err:
        click.echo(f"ERROR: {err}", err=True)
        sys.exit(1)


@cli.command()
@click.argument("base_path", type=Path)
@click.option("--host", default="127.0.0.1")
@click.option("--port", default=1337)
@click.option("--open", is_flag=True, help="Open page in browser")
def serve(base_path, host, port, open):
    """
    Start a web-server to view the documentation
    """
    build_path = base_path / "docs" / "_build"
    try:
        process = subprocess.Popen(
            [
                "python",
                "-m",
                "http.server",
                "--directory",
                str(build_path),
                "--bind",
                host,
                f"{port}",
            ],
            universal_newlines=True,
        )
        if open:
            url = f"http://{host}:{port}/"
            built_docs = [
                b.name
                for b in build_path.iterdir()
                if b.is_dir() and not b.name.startswith(".") and b.name != "doctest"
            ]
            if len(built_docs) == 1:
                # Only one package, dir-listing not necessary.
                url = f"{url}{built_docs[0]}/latest"

            click.launch(url)

        process.wait()

    except KeyboardInterrupt:
        click.echo("Server interrupted by user.")

    except Exception as e:
        click.echo(f"An error occurred: {e}", err=True)


@cli.command()
@click.argument("base_path", type=Path)
def test(base_path):
    """
    Run doc-tests
    """
    tester = _create_tester(base_path=base_path)
    try:
        tester.test()
    except ApeDocsTestError as err:
        click.echo(f"ERROR: {err}", err=True)
        sys.exit(1)


@cli.command()
@click.argument("base_path", type=Path)
@build_mode_option()
@click.option(
    "--repo",
    "--repository",
    help="Specify the repository ID",
)
@click.option(
    "--skip-push",
    is_flag=True,
    hidden=True,
)
def publish(base_path, mode, repo, skip_push):
    """
    Publish docs
    """
    builder = _create_builder(mode=mode, base_path=base_path)
    try:
        builder.publish(repository=repo, push=not skip_push)
    except ApeDocsPublishError as err:
        click.echo(f"ERROR: {err}", err=True)
        sys.exit(1)


@cli.command()
@click.argument("base_path", type=Path)
def clean(base_path):
    """
    Delete build artifacts
    """
    builder = _create_builder(base_path=base_path)
    builder.clean()


def _create_builder(*args, **kwargs) -> DocumentationBuilder:
    # Abstracted for testing purposes.
    return DocumentationBuilder(*args, **kwargs)


def _create_tester(*args, **kwargs) -> DocumentationTester:
    # Abstracted for testing purposes.
    return DocumentationTester(*args, **kwargs)



import subprocess
from pathlib import Path

from sphinx_ape._base import Documentation
from sphinx_ape.exceptions import ApeDocsBuildError, ApeDocsTestError


class DocumentationTester(Documentation):
    """
    Small wrapper around sphinx-build's doctest command.
    """

    @property
    def doctest_folder(self) -> Path:
        """
        The path to doctest's build folder.
        """
        return self.build_path.parent / "doctest"

    @property
    def doctest_output_file(self) -> Path:
        """
        The path to doctest's output file.
        """
        return self.doctest_folder / "output.txt"

    def test(self):
        """
        Run the sphinx-build doctest command.

        Raises:
            :class:`~sphinx_ape.exceptions.ApeDocsTestError`
        """
        self._run_tests()
        output = self.doctest_output_file.read_text() if self.doctest_output_file.is_file() else ""
        if "0 failed" in output or "0 tests" in output:
            # Either no failures or no tests ran.
            return

        # Failures.
        raise ApeDocsTestError(output)

    def _run_tests(self):
        try:
            subprocess.run(
                ["sphinx-build", "-b", "doctest", "docs", str(self.doctest_folder)], check=True
            )
        except subprocess.CalledProcessError as err:
            raise ApeDocsBuildError(str(err)) from err



from pathlib import Path

from docutils.parsers.rst import directives
from sphinx.util.docutils import SphinxDirective

from sphinx_ape.build import DocumentationBuilder


class DynamicTocTree(SphinxDirective):
    """
    Dynamically create the TOC-tree so users don't
    need to create and maintain index.html files.
    """

    option_spec = {
        "title": directives.unchanged,
    }

    @property
    def _base_path(self) -> Path:
        env = self.state.document.settings.env
        return Path(env.srcdir)

    @property
    def title(self) -> str:
        if res := self.options.get("title"):
            # User configured the title.
            return res

        # Deduced: "Ape-Docs" or "Ape-Vyper-Docs", etc.
        name = self._base_path.parent.name
        name_parts = [n.capitalize() for n in name.split("-")]
        capped_name = "-".join(name_parts)
        return f"{capped_name}-Docs"

    @property
    def _title_rst(self) -> str:
        title = self.title
        bar = "=" * len(title)
        return f"{title}\n{bar}"

    @property
    def builder(self) -> DocumentationBuilder:
        return DocumentationBuilder(base_path=self._base_path.parent)

    def run(self):
        userguides = self._get_userguides()
        cli_docs = self._get_cli_references()
        methoddocs = self._get_methoddocs()
        plugin_methoddocs = [d for d in methoddocs if d.startswith("ape-")]
        methoddocs = [d for d in methoddocs if d not in plugin_methoddocs]
        sections = {"User Guides": userguides, "CLI Reference": cli_docs}
        if plugin_methoddocs:
            # Core (or alike)
            sections["Plugin Python Reference"] = plugin_methoddocs
            sections["Core Python Reference"] = methoddocs
        else:
            # Plugin or regular package.
            sections["Python Reference"] = methoddocs

        toc_trees = []
        for caption, entries in sections.items():
            if len(entries) < 1:
                continue

            toc_tree = f".. toctree::\n   :caption: {caption}\n   :maxdepth: 1\n\n"
            for entry in entries:
                toc_tree += f"   {entry}\n"

            toc_trees.append(toc_tree)

        toc_tree_rst = "\n".join(toc_trees)
        restructured_text = self._title_rst
        if toc_tree_rst:
            restructured_text = f"{restructured_text}\n\n{toc_tree_rst}"

        return self.parse_text_to_nodes(restructured_text)

    def _get_userguides(self) -> list[str]:
        return [f"userguides/{n}" for n in self.builder.userguide_names]

    def _get_cli_references(self) -> list[str]:
        return [f"commands/{n}" for n in self.builder.cli_reference_names]

    def _get_methoddocs(self) -> list[str]:
        return [f"methoddocs/{n}" for n in self.builder.methoddoc_names]



import os
import shutil
from enum import Enum
from pathlib import Path
from typing import Optional, Union

from sphinx_ape._base import Documentation
from sphinx_ape._utils import extract_source_url, git, replace_tree, sphinx_build
from sphinx_ape.exceptions import ApeDocsBuildError, ApeDocsPublishError

REDIRECT_HTML = """
<!DOCTYPE html>
<meta charset="utf-8">
<title>Redirecting...</title>
<meta http-equiv="refresh" content="0; URL=./{}/">
"""


class BuildMode(Enum):
    LATEST = 0
    """Build and then push to 'latest/'"""

    MERGE_TO_MAIN = 1
    """Build and then push to 'stable/'"""

    RELEASE = 2
    """Build and then push to 'stable/', 'latest/', and the version's release tag folder"""

    @classmethod
    def init(cls, identifier: Optional[Union[str, "BuildMode"]] = None) -> "BuildMode":
        if identifier is None:
            # Default.
            return BuildMode.LATEST

        elif isinstance(identifier, BuildMode):
            return identifier

        elif isinstance(identifier, int):
            return BuildMode(identifier)

        elif isinstance(identifier, str):
            if "." in identifier:
                # Click being weird, value like "buildmode.release".
                identifier = identifier.split(".")[-1].upper()

            identifier = identifier.lower()
            if identifier == "release":
                return BuildMode.RELEASE
            elif identifier in ("push", "merge_to_main"):
                return BuildMode.MERGE_TO_MAIN
            else:
                return BuildMode.LATEST

        # Unexpected.
        raise TypeError(identifier)


class DocumentationBuilder(Documentation):
    """
    Builds either "latest", or "stable" / "release"
    documentation.
    """

    def __init__(
        self,
        mode: Optional[BuildMode] = None,
        base_path: Optional[Path] = None,
        name: Optional[str] = None,
        pages_branch_name: Optional[str] = None,
    ) -> None:
        self.mode = BuildMode.LATEST if mode is None else mode
        super().__init__(base_path, name)
        self._pages_branch_name = pages_branch_name or "gh-pages"

    def build(self):
        """
        Build the documentation.

        Example:
            >>> builder = DocumentationBuilder(
            ...   mode=BuildMode.LATEST,
            ...   base_path=Path.cwd(),
            ...   name="sphinx-ape"
            ... )
            >>> builder.build()

        Raises:
            :class:`~sphinx_ape.exceptions.ApeDocsBuildError`: When
              building fails.
        """

        if self.mode in (BuildMode.LATEST, BuildMode.MERGE_TO_MAIN):
            # TRIGGER: Push to 'main' branch. Only builds latest.
            #   And on PRs / local.
            self._sphinx_build(self.latest_path)

        elif self.mode is BuildMode.RELEASE:
            # TRIGGER: Release on GitHub
            self._build_release()

        else:
            # Unknown 'mode'.
            raise ApeDocsBuildError(f"Unsupported build-mode: {self.mode}")

        self._setup_redirect()

    def clean(self):
        shutil.rmtree(self.root_build_path)

    def publish(self, repository: Optional[str] = None, push: bool = True):
        """
        Publish the documentation to GitHub pages.
        Meant to be run in CI/CD on releases.

        Args:
            repository (Optional[str]]): The repository name. Defaults to GitHub env-var
              or extraction from setup file.
            push (bool): Set to ``False`` to skip git add, commit, and push.

        Raises:
            :class:`~sphinx_ape.exceptions.ApeDocsPublishError`: When
              publishing fails.
        """
        try:
            self._publish(repository=repository, push=push)
        except Exception as err:
            raise ApeDocsPublishError(str(err)) from err

    def _publish(self, repository: Optional[str] = None, push: bool = True):
        if repository:
            repo_url = f"https://github.com/{repository}"
        else:
            repo_url = extract_source_url()

        gh_pages_path = self._base_path / "gh-pages"
        git(
            "clone",
            repo_url,
            "--branch",
            self._pages_branch_name,
            "--single-branch",
            self._pages_branch_name,
        )
        try:
            # Any built docs get added; the docs that got built are based on
            # the mode parameter.
            for path in self.build_path.iterdir():
                if path.is_dir() and not path.name.startswith(".") and path.name != "doctest":
                    dst_path = gh_pages_path / path.name
                    if dst_path.is_dir():
                        shutil.rmtree(dst_path)

                    shutil.copytree(path, dst_path)

                elif (path.name == "index.html") and path.is_file():
                    gh_pages_path.mkdir(exist_ok=True)
                    (gh_pages_path / "index.html").write_text(path.read_text())

            no_jykell_file = gh_pages_path / ".nojekyll"
            no_jykell_file.touch(exist_ok=True)

            if push:
                here = os.getcwd()
                try:
                    os.chdir(str(gh_pages_path))
                    # NOTE: CI/CD does not push here but instead uses the
                    #  push-action w/ a login.
                    git("add", ".")
                    git("commit", "-m", "Update documentation", "-a")
                    git("push", "origin", "gh-pages")
                finally:
                    os.chdir(here)

        finally:
            if push:
                # Only delete if we are done pushing.
                # Else, leave so the GH action can push.
                shutil.rmtree(gh_pages_path, ignore_errors=True)

    def _build_release(self):
        if not (tag := git("describe", "--tag")):
            raise ApeDocsBuildError("Unable to find release tag.")

        if "beta" in tag or "alpha" in tag:
            # Avoid creating release directory for beta
            # or alpha releases. Only update "stable" and "latest".
            self._sphinx_build(self.stable_path)
            replace_tree(self.stable_path, self.latest_path)

        else:
            # Use the tag to create a new release folder.
            build_dir = self.build_path / tag
            self._sphinx_build(build_dir)

            if not build_dir.is_dir():
                return

            # Clean-up unnecessary extra 'fonts/' directories to save space.
            # There should still be one in 'latest/'
            for font_dirs in build_dir.glob("**/fonts"):
                if font_dirs.is_dir():
                    shutil.rmtree(font_dirs)

            # Replace 'stable' and 'latest' with this version.
            for path in (self.stable_path, self.latest_path):
                replace_tree(build_dir, path)

    def _setup_redirect(self):
        self.build_path.mkdir(exist_ok=True, parents=True)
        redirect = "stable" if self.stable_path.is_dir() else "latest"
        # We replace it to handle the case when stable has joined the chat.
        self.index_file.unlink(missing_ok=True)
        self.index_file.write_text(REDIRECT_HTML.format(redirect))

    def _sphinx_build(self, dst_path: Path):
        shutil.rmtree(dst_path, ignore_errors=True)
        sphinx_build(dst_path, self.docs_path)



from pathlib import Path

import pytest

from sphinx_ape._utils import extract_package_name, extract_source_url


@pytest.fixture(scope="session")
def create_setup_py():
    def fn(
        base_path: Path, name: str = "ape-myplugin", url: bool = False, source_url: bool = False
    ):
        content = f"""
#!/usr/bin/env python

from setuptools import find_packages, setup
extras_require = {{
    "test": [  # `test` GitHub Action jobs uses this
        "pytest>=6.0",  # Core testing package
    ],
}}
setup(
    name="{name}",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.12",
    ],
"""
        if url:
            content = f'{content}\n  url="https://github.com/ApeWorX/{name}",'
        if source_url:
            content = (
                f"{content}    project_urls={{\n        "
                f'"Source": "https://github.com/ApeWorX/{name}",\n    }},'
            )

        content = f"{content}\n)"
        path = base_path / "setup.py"
        path.write_text(content)
        return path

    return fn


def test_extract_package_name_setup_py(temp_path, create_setup_py):
    name = "ape-myplugin"
    create_setup_py(temp_path, name)
    actual = extract_package_name(temp_path)
    assert actual == name


def test_extract_source_url_from_setup_py_project_urls_source(temp_path, create_setup_py):
    name = "ape-myplugin"
    create_setup_py(temp_path, name, source_url=True)
    actual = extract_source_url(temp_path)
    expected = f"https://github.com/ApeWorX/{name}"
    assert actual == expected


def test_extract_source_url_from_setup_py_url(temp_path, create_setup_py):
    name = "ape-myplugin"
    create_setup_py(temp_path, name, url=True)
    actual = extract_source_url(temp_path)
    expected = f"https://github.com/ApeWorX/{name}"
    assert actual == expected



from pathlib import Path

from sphinx_ape.testing import DocumentationTester


class TestDocumentationTester:
    def test_test(self, temp_path):
        tester = DocumentationTester(base_path=Path(__file__).parent.parent)
        tester.test()
        assert tester.doctest_output_file.is_file()



from pathlib import Path

import pytest

from sphinx_ape.build import REDIRECT_HTML, BuildMode, DocumentationBuilder


class TestBuildMode:
    @pytest.mark.parametrize("val", ("latest", 0, "pull_request", "buildmode.latest"))
    def test_init_latest(self, val):
        mode = BuildMode.init(val)
        assert mode is BuildMode.LATEST

    @pytest.mark.parametrize("val", ("merge_to_main", 1, "push", "buildmode.merge_to_main"))
    def test_init_merge_to_main(self, val):
        mode = BuildMode.init(val)
        assert mode is BuildMode.MERGE_TO_MAIN, val

    @pytest.mark.parametrize("val", ("release", 2, "buildmode.release"))
    def test_init_release(self, val):
        mode = BuildMode.init(val)
        assert mode is BuildMode.RELEASE


def assert_build_path(path: Path, expected: str):
    assert path.name == expected  # Tag, latest, or stable
    assert path.parent.name == "sphinx-ape"  # Project name, happens to be this lib
    assert path.parent.parent.name == "_build"  # Sphinx-ism
    assert path.parent.parent.parent.name == "docs"  # Sphinx-ism


class TestDocumentationBuilder:
    @pytest.fixture
    def mock_sphinx(self, mocker):
        def run_mock_sphinx(path, *args, **kwargs):
            path.mkdir(parents=True)
            buildfile = path / "build.txt"
            buildfile.touch()

        mock = mocker.patch("sphinx_ape.build.sphinx_build")
        mock.side_effect = run_mock_sphinx
        return mock

    @pytest.fixture(autouse=True)
    def mock_git(self, mocker):
        return mocker.patch("sphinx_ape.build.git")

    def test_build_latest(self, temp_path):
        builder = DocumentationBuilder(mode=BuildMode.LATEST, base_path=temp_path)
        builder.init()  # so there is something to build.
        builder.build()
        assert builder.latest_path.is_dir()

        # Ensure re-direct exists and points to latest/.
        assert builder.index_file.is_file()
        expected_content = REDIRECT_HTML.format("latest")
        assert builder.index_file.read_text() == expected_content

        # Ensure static content exists.
        assert (builder.latest_path / "_static").is_dir()
        assert (builder.latest_path / "_static" / "logo_green.svg").is_file()

    def test_build_release(self, mock_sphinx, mock_git, temp_path):
        tag = "v1.0.0"
        mock_git.return_value = tag
        builder = DocumentationBuilder(mode=BuildMode.RELEASE, base_path=temp_path)
        builder.build()
        call_path = mock_sphinx.call_args[0][0]
        assert_build_path(call_path, tag)
        # Latest and Stable should also have been created!
        assert_build_path(call_path.parent / "latest", "latest")
        assert_build_path(call_path.parent / "stable", "stable")
        # Ensure re-direct exists and points to stable/.
        assert builder.index_file.is_file()
        expected_content = REDIRECT_HTML.format("stable")
        assert builder.index_file.read_text() == expected_content

    @pytest.mark.parametrize("sub_tag", ("alpha", "beta"))
    def test_build_alpha_release(self, sub_tag, mock_sphinx, mock_git, temp_path):
        """
        We don't build version releases when using alpha or beta, but we
        still update "stable" and "latest".
        """
        tag = f"v1.0.0{sub_tag}"
        mock_git.return_value = tag
        builder = DocumentationBuilder(mode=BuildMode.RELEASE, base_path=temp_path)
        builder.build()
        call_path = mock_sphinx.call_args[0][0]
        assert_build_path(call_path, "stable")
        # Latest should also have been created!
        assert_build_path(call_path.parent / "latest", "latest")

    def test_publish_merge_to_main(self, temp_path, mock_git):
        tag = "v1.0.0"
        mock_git.return_value = tag
        builder = DocumentationBuilder(mode=BuildMode.MERGE_TO_MAIN, base_path=temp_path)
        gh_pages_path = temp_path / "gh-pages"
        nojekyll_file = gh_pages_path / ".nojekyll"
        stable_dir = gh_pages_path / "stable"
        latest_dir = gh_pages_path / "latest"
        tag_dir = gh_pages_path / tag
        index_file = gh_pages_path / builder.index_file.name
        static_dir = latest_dir / "_static"

        # Create a random file in _static to show it doesn't matter.
        static_dir.mkdir(exist_ok=True, parents=True)
        random_file = static_dir / "randomfile.txt"
        random_file.write_text("this should be fine.")

        # Ensure built first.
        builder.init()
        builder.build()
        builder.publish(push=False)

        assert gh_pages_path.is_dir()
        assert nojekyll_file.is_file()
        assert latest_dir.is_dir()
        assert index_file.is_file()
        # Not tag-release should have gotten created on merge-to-main.
        assert not tag_dir.is_dir()
        # Stable only gets built on releases.
        assert not stable_dir.is_dir()
        # Ensure static content exists.
        assert static_dir.is_dir(), "Missing 'latest/_static'"
        logo = static_dir / "logo_green.svg"
        assert logo.is_file(), "Missing logo: 'latest/_static/logo_green.svg'"
        assert not random_file.is_file()

    def test_publish_release(self, temp_path, mock_git):
        tag = "v1.0.0"
        mock_git.return_value = tag
        builder = DocumentationBuilder(mode=BuildMode.RELEASE, base_path=temp_path)
        builder.init()
        # Ensure built first.
        builder.build()
        builder.publish(push=False)
        gh_pages_path = temp_path / "gh-pages"
        nojekyll_file = gh_pages_path / ".nojekyll"
        stable_dir = gh_pages_path / "stable"
        latest_dir = gh_pages_path / "latest"
        tag_dir = gh_pages_path / tag
        index_file = gh_pages_path / builder.index_file.name
        assert gh_pages_path.is_dir()
        assert nojekyll_file.is_file()
        assert stable_dir.is_dir()
        assert latest_dir.is_dir()
        assert tag_dir.is_dir()
        assert index_file.is_file()
        # Ensure static content exists.
        for directory in (latest_dir, stable_dir, tag_dir):
            static_dir = directory / "_static"
            assert static_dir.is_dir(), f"Missing static: {directory.name}"
            logo = static_dir / "logo_green.svg"
            assert logo.is_file(), f"Missing logo: {directory.name}"



from pathlib import Path
from typing import Optional

from sphinx_ape._utils import get_package_name


class Documentation:
    def __init__(self, base_path: Optional[Path] = None, name: Optional[str] = None) -> None:
        self._base_path = base_path or Path.cwd()
        self._name = name or get_package_name()

    @property
    def docs_path(self) -> Path:
        return self._base_path / "docs"

    @property
    def root_build_path(self) -> Path:
        return self.docs_path / "_build"

    @property
    def build_path(self) -> Path:
        return self.root_build_path / self._name

    @property
    def latest_path(self) -> Path:
        return self.build_path / "latest"

    @property
    def stable_path(self) -> Path:
        return self.build_path / "stable"

    @property
    def userguides_path(self) -> Path:
        return self.docs_path / "userguides"

    @property
    def commands_path(self) -> Path:
        return self.docs_path / "commands"

    @property
    def methoddocs_path(self) -> Path:
        return self.docs_path / "methoddocs"

    @property
    def conf_file(self) -> Path:
        return self.docs_path / "conf.py"

    @property
    def index_file(self) -> Path:
        return self.build_path / "index.html"

    def init(self):
        if not self.docs_path.is_dir():
            self.docs_path.mkdir()

        self._ensure_quickstart_exists()
        self._ensure_conf_exists()
        self._ensure_index_exists()

    def _ensure_conf_exists(self):
        if self.conf_file.is_file():
            return

        content = 'extensions = ["sphinx_ape"]\n'
        self.conf_file.write_text(content)

    def _ensure_index_exists(self):
        index_file = self.docs_path / "index.rst"
        if index_file.is_file():
            return

        content = ".. dynamic-toc-tree::\n"
        index_file.write_text(content)

    def _ensure_quickstart_exists(self):
        quickstart_path = self.userguides_path / "quickstart.md"
        if quickstart_path.is_file():
            # Already exists.
            return

        self.userguides_path.mkdir(exist_ok=True)
        quickstart_path.write_text("```{include} ../../README.md\n```\n")

    @property
    def userguide_names(self) -> list[str]:
        guides = self._get_filenames(self.userguides_path)
        quickstart_name = "userguides/quickstart"
        if quickstart_name in guides:
            # Make sure quick start is first.
            guides = [quickstart_name, *[g for g in guides if g != quickstart_name]]

        return guides

    @property
    def cli_reference_names(self) -> list[str]:
        return self._get_filenames(self.commands_path)

    @property
    def methoddoc_names(self) -> list[str]:
        return self._get_filenames(self.methoddocs_path)

    def _get_filenames(self, path: Path) -> list[str]:
        if not path.is_dir():
            return []

        return sorted([g.stem for g in path.iterdir() if g.suffix in (".md", ".rst")])



from pathlib import Path

import pytest
from click.testing import CliRunner

from sphinx_ape._cli import cli as root_cli


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def cli():
    return root_cli


@pytest.fixture(autouse=True)
def mock_builder(mocker):
    patch = mocker.patch("sphinx_ape._cli._create_builder")
    builder = mocker.MagicMock()
    patch.return_value = builder
    return builder


def test_build(runner, cli):
    result = runner.invoke(cli, ("build", "."))
    assert result.exit_code == 0
    assert "Building 'sphinx-ape' 'LATEST'" in result.output


def test_build_docs_folder_not_exists(runner, cli, mock_builder):
    mock_builder.docs_path = Path("thisdoesnotexist")
    result = runner.invoke(cli, ("build", "."))
    assert result.exit_code != 0
    assert "docs/ folder missing. Try running:\n\tsphinx-ape init" in result.output


def test_build_from_pr(runner, cli):
    result = runner.invoke(cli, ("build", ".", "--mode", "pull_request"))
    assert result.exit_code == 0
    assert "Building 'sphinx-ape' 'LATEST'" in result.output


def test_publish(runner, cli):
    result = runner.invoke(cli, ("publish", ".", "--skip-push"))
    assert result.exit_code == 0



class SphinxApeException(Exception):
    """
    Base exception.
    """


class ApeDocsBuildError(SphinxApeException):
    """
    Building the docs failed.
    """


class ApeDocsTestError(SphinxApeException):
    """
    Running doc-tests failed.
    """


class ApeDocsPublishError(SphinxApeException):
    """
    Publishing the docs failed.
    """



import os
import sys
from pathlib import Path

from sphinx.application import Sphinx
from sphinx.util import logging

from sphinx_ape._utils import get_package_name
from sphinx_ape.sphinx_ext.directives import DynamicTocTree

logger = logging.getLogger(__name__)


def setup(app: Sphinx):
    """Set default values for various Sphinx configurations."""

    # For building and serving multiple projects at once,
    # we situate ourselves in the parent directory.
    sys.path.insert(0, os.path.abspath(".."))

    # Configure project and other one-off items.
    package_name = get_package_name()
    app.config.project = package_name
    app.config.copyright = "2024, ApeWorX LTD"
    app.config.author = "ApeWorX Team"

    app.config.exclude_patterns = list(
        set(app.config.exclude_patterns).union({"_build", ".DS_Store"})
    )
    app.config.source_suffix = [".rst", ".md"]
    app.config.master_doc = "index"

    # Automatically add required extensions.
    default_extensions = {
        "myst_parser",
        "sphinx_click",
        "sphinx.ext.autodoc",
        "sphinx.ext.autosummary",
        "sphinx.ext.doctest",
        "sphinx.ext.napoleon",
        "sphinx_rtd_theme",
        "sphinx_plausible",
    }

    # Ensure these have been loaded before continuing.
    for ext in default_extensions:
        app.setup_extension(ext)

    app.config.extensions = list(set(app.config.extensions).union(default_extensions))

    # Plausible config.
    if not getattr(app.config, "plausible_domain", None):
        app.config.plausible_domain = "docs.apeworx.io"

    # Configure the HTML workings.
    static_dir = Path(__file__).parent.parent / "_static"

    app.config.html_theme = "shibuya"
    app.config.html_favicon = str(static_dir / "favicon.ico")
    app.config.html_baseurl = package_name
    app.config.html_static_path = [str(static_dir)]
    app.config.html_theme_options = {
        "light_logo": "_static/logo_grey.svg",
        "dark_logo": "_static/logo_green.svg",
        "accent_color": "lime",
    }

    # All MyST config.
    app.config.myst_all_links_external = True

    # Any config starting with "auto".
    app.config.autosummary_generate = True
    exclude_members = (
        # Object.
        "__weakref__",
        "__metaclass__",
        "__init__",
        "__format__",
        "__new__",
        "__dir__",
        # Pydantic.
        "model_config",
        "model_fields",
        "model_post_init",
        "model_computed_fields",
        "__class_vars__",
        "__private_attributes__",
        "__pydantic_complete__",
        "__pydantic_core_schema__",
        "__pydantic_custom_init__",
        "__pydantic_decorators__",
        "__pydantic_generic_metadata__",
        "__pydantic_parent_namespace__",
        "__pydantic_post_init__",
        "__pydantic_serializer__",
        "__pydantic_validator__",
        # EthPydanticTypes.
        "__ape_extra_attributes__",
    )
    app.config.autodoc_default_options = {"exclude-members": ", ".join(exclude_members)}

    # Add the directive enabling the dynamic-TOC.
    app.add_directive("dynamic-toc-tree", DynamicTocTree)

    # Output data needed for the rest of the build.
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



import ast
import os
import shutil
import subprocess
from pathlib import Path
from typing import Any, Optional, Union

import tomli

from sphinx_ape.exceptions import ApeDocsBuildError

# Avoid needing to have common Ape packages re-configure this.
PACKAGE_ALIASES = {
    "eth-ape": "ape",
}


def git(*args):
    return subprocess.check_output(["git", *args]).decode("ascii").strip()


def new_dir(path: Path) -> Path:
    if path.is_dir():
        shutil.rmtree(path)

    path.mkdir(parents=True)
    return path


def sphinx_build(dst_path: Path, source_dir: Union[Path, str]) -> Path:
    path = new_dir(dst_path)
    try:
        subprocess.check_call(["sphinx-build", str(source_dir), str(path)])
    except subprocess.SubprocessError as err:
        raise ApeDocsBuildError(f"Command 'sphinx-build docs {path}' failed.") from err

    return path


def get_source_url(directory: Optional[Path] = None) -> str:
    if env_var := os.getenv("GITHUB_REPO"):
        return f"https://github.com/{env_var}"

    return extract_source_url(directory=directory)


def extract_source_url(directory: Optional[Path] = None) -> str:
    directory = directory or Path.cwd()
    url = None
    if (directory / "setup.py").is_file():
        url = _extract_github_url_from_setup_py(directory / "setup.py")
    if url is None:
        raise ApeDocsBuildError("No package source URL found.")

    return url


def _extract_github_url_from_setup_py(file_path: Path) -> Optional[str]:
    # Check `project_urls`
    project_urls: dict = _extract_key_from_setup_py("project_urls", file_path) or {}  # type: ignore
    if url := project_urls.get("Source"):
        return url

    # Try url
    url = _extract_key_from_setup_py("url", file_path)
    if url and url.startswith("https://github.com"):
        return url

    return None


def _extract_name_from_setup_py(file_path: Path) -> Optional[str]:
    return _extract_key_from_setup_py("name", file_path)


def _extract_key_from_setup_py(key: str, file_path: Path) -> Optional[Any]:
    if not (setup_content := file_path.read_text()):
        return None
    elif not (parsed_content := ast.parse(setup_content)):
        return None

    # Walk through the AST to find the setup() call and extract the desired information
    for node in ast.walk(parsed_content):
        if (
            not isinstance(node, ast.Call)
            or not hasattr(node.func, "id")
            or node.func.id != "setup"
        ):
            continue

        for keyword in node.keywords:
            if keyword.arg != key:
                continue

            return ast.literal_eval(keyword.value)

    return None


def _extract_name_from_pyproject_toml(file_path: Path) -> Optional[str]:
    """Extract package name from pyproject.toml."""
    with open(file_path, "rb") as file:
        pyproject = tomli.load(file)

    if "tool" in pyproject and "poetry" in pyproject["tool"]:
        return pyproject["tool"]["poetry"].get("name")

    elif "project" in pyproject:
        return pyproject["project"].get("name")

    return None


def get_package_name() -> str:
    if env_var := os.getenv("GITHUB_REPO"):
        return env_var

    # Figure it out.
    return extract_package_name()


def extract_package_name(directory: Optional[Path] = None) -> str:
    """Detect and extract the package name from the project files."""
    directory = directory or Path.cwd()
    pkg_name = None
    if (directory / "setup.py").is_file():
        pkg_name = _extract_name_from_setup_py(directory / "setup.py")
    if pkg_name is None and (directory / "pyproject.toml").is_file():
        pkg_name = _extract_name_from_pyproject_toml(directory / "pyproject.toml")
    if pkg_name is None:
        path = f"{directory}".replace(f"{Path.home()}", "$HOME")
        raise ApeDocsBuildError(f"No package name found at '{path}'.")

    return PACKAGE_ALIASES.get(pkg_name, pkg_name)


def replace_tree(base_path: Path, dst_path: Path):
    shutil.rmtree(dst_path, ignore_errors=True)
    shutil.copytree(base_path, dst_path)
